/*
 * Copyright (C) 2014 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <ctype.h>                      /* isspace() */
#include <debug.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "vector.h"


#define PARAMS_MAX 4096
#define COLORS_MAX 16

#ifndef MAX
# define MAX(a,b) ((a)>(b)?(a):(b))
#endif

/* --- utility functions ------------------------------------------------ */

static int
_atoin(const char* s,size_t len)
{
    const char *end = s+len;
    int neg;
    int res;
    while(s<end && isspace(*s))
        ++s;
    if(s>=end)
        return 0;
    if('-'==*s){
        neg = 1;
        ++s;
    }else{
        neg = 0;
    }
    res = 0;
    while(s<end && *s>='0' && *s<='9'){
        res *= 10;
        res += *s-'0';
        ++s;
    }
    return neg ? -res : res;
}


static const char *
_strnchrnul(const char *s,size_t len,int c)
{
    const char *end = s+len;
    for(;s<end && *s!=c;++s);
    return s;
}


static void
_memcycle(void *mem, size_t length, off_t offset)
{
    char buffer[4096];
    if(offset<0){
        offset = length - (-offset)%length;
    }else{
        offset = offset%length;
    }
    if(offset < length/2){
        memcpy(buffer,mem+length-offset,offset);
        memmove(mem+offset,mem,length-offset);
        memcpy(mem,buffer,offset);
    }else{
        memcpy(buffer,mem,length-offset);
        memmove(mem,mem+length-offset,offset);
        memcpy(mem+offset,buffer,length-offset);
    }
}

/* --- str_t ------------------------------------------------------------ */

typedef struct{
    const char *str;
    size_t len;
}str_t;

static void
_str_adv(str_t *sn, size_t bytes)
{
    if(sn->len<bytes)
        bytes = sn->len;
    sn->str += bytes;
    sn->len -= bytes;
}

#define _str_toi(sn) _atoin((sn)->str,(sn)->len)
#define _str_cmp(sn,s) strncmp((sn)->str,s,(sn)->len)

/* ---------------------------------------------------------------------- */

typedef struct{
    double r;
    double g;
    double b;
}color_t;

struct _shape_t{
    enum{
        VECTOR_RECT,
        VECTOR_PATH
    }type;
    vector_shape_style_t style;
    int color;
    union{
        struct{
            int point;
            int width;
            int height;
        }rect;
        vector_path_t path;
    };
};

struct _anim_t{
    enum{
        AT_BALANCE,
        AT_CYCLE_COLORS,
        AT_MOVE_POINT,
        AT_MOVE_SHAPE,
        AT_PATH_SHAPE,
        AT_ROTATE,
        AT_SIZE
    }type;
    union{
        struct{
            int shape;
            int point;
            int prev;
            int next;
        }balance;
        struct{
            int direction;
            int first;
            int count;
            int steps;
        }cycle_colors;
        struct{
            int first;
            int last;
            int dx;
            int dy;
        }move_point;
        struct{
            int first;
            int last;
            int dx;
            int dy;
        }move_shape;
        struct{
            int first;
            int last;
            int steps;
            int point;
        }path_shape;
        struct{
            int first;
            int last;
            int cx;
            int cy;
            double rx;
            double ry;
            double rz;
            double angle;
        }rotate;
        struct{
            int first;
            int last;
            int cx;
            int cy;
            double scale;
        }size;
    };
};

struct _keyframe_t{
    double start;
    double duration;
    color_t color[COLORS_MAX];
    point_t *point;
    int point_count;
    vector_segment_t *segment;
    int segment_count;
    shape_t *shape;
    int shape_count;
    anim_t *anim;
    int anim_count;
};

typedef struct{
    double xx;
    double xy;
    double yx;
    double yy;
    double zx;
    double zy;
}rotation_t;

typedef struct {
    int point;
    int shape;
}killpoint_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    keyframe_t *keyframe;
};

typedef struct{
    str_t param[PARAMS_MAX];
    int param_count;
}command_t;


static const char **
_anist_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-ani-st",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_anist_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->pixel_ratio = 5.0/6;
    h->info->scalable_size = 1;
    h->info->scalable_time = 1;
    h->info->threadsafe = 1;
    h->keyframe = NULL;
    return h;
}


static void
_anist_free(abydos_plugin_handle_t *h)
{
    if(h->keyframe){
        int i;
        for(i=0;i<h->info->frame_count;++i){
            keyframe_t *kf = &h->keyframe[i];
            if(kf->point)
                free(kf->point);
            if(kf->segment)
                free(kf->segment);
            if(kf->shape)
                free(kf->shape);
            if(kf->anim)
                free(kf->anim);
        }
        free(h->keyframe);
    }
    free(h);
}



static int
_read_token(str_t *token, str_t *sn)
{
    while(sn->len > 0 && isspace(sn->str[0]))
        _str_adv(sn,1);
    if(sn->len <= 0)
        return -1;

    if('('==sn->str[0]){
        const char *end;
        _str_adv(sn,1);
        token->str = sn->str;
        end = _strnchrnul(sn->str,sn->len,')');
        token->len = end-token->str;
        _str_adv(sn,token->len);
        if(')'==*end)
            _str_adv(sn,1);
        return 0;
    }else{
        token->str = sn->str;
        while(sn->len > 0){
            _str_adv(sn,1);
            if(isspace(sn->str[0]))
                break;
        }
        token->len = sn->str - token->str;
    }
    return 0;
}


static int
_read_command(command_t *cmd, str_t *sn)
{
    str_t line;
    while(sn->len > 0 && sn->str[0] != '*')
        _str_adv(sn,1);
    if(sn->len <= 1)
        return -1;
    _str_adv(sn,1);
    line.str = sn->str;
    line.len = _strnchrnul(sn->str,sn->len,'*')-line.str;
    _str_adv(sn,line.len);

    cmd->param_count = 0;
    while(line.len>0 && cmd->param_count < PARAMS_MAX){
        if(_read_token(&cmd->param[cmd->param_count],&line)<0)
            break;
        ++cmd->param_count;
    }
    return 0;
}


static void
_read_color(color_t *c, str_t *sn){
    str_t token;
    c->r = _read_token(&token,sn) ? 0 : (double)_str_toi(&token)/255;
    c->g = _read_token(&token,sn) ? 0 : (double)_str_toi(&token)/255;
    c->b = _read_token(&token,sn) ? 0 : (double)_str_toi(&token)/255;
    DEBUG_p("%f %f %f - %d %d %d",c->r,c->g,c->b,(int)(c->r*255),(int)(c->g*255),(int)(c->b*255));
}


static void
_read_point(point_t *p, str_t *sn){
    str_t token;
    p->x = _read_token(&token,sn) ? 0 : _str_toi(&token);
    p->y = _read_token(&token,sn) ? 0 : _str_toi(&token);
}


static void
_set_default_palette(color_t *c)
{
    static const int palette[COLORS_MAX] = {
        0x0000, 0x0700, 0x0750, 0x0670, 0x0070, 0x0056, 0x0437, 0x0607,
        0x0530, 0x0433, 0x0141, 0x0652, 0x0334, 0x0445, 0x0556, 0x0777
    };
    int i;
    for(i=0;i<COLORS_MAX;++i){
        c[i].r = (double)((palette[i]>>8)&7)/7;
        c[i].g = (double)((palette[i]>>4)&7)/7;
        c[i].b = (double)((palette[i]>>0)&7)/7;
    }
}


static void
_shape_points(const shape_t *shape, const vector_segment_t *seg, int *first, int *last)
{
    switch(shape->type){
    case VECTOR_RECT:
        *first = *last = shape->rect.point;
        break;
    case VECTOR_PATH:
        *first = seg[shape->path.segment].point;
        *last = *first + shape->path.point_count -1;
        break;
    };
}


static void
_debug_points(vector_buffer_t *vb)
{
    point_t *p = vector_buffer_get_point(vb, 0);
    int i;
    for(i=0;i<vector_buffer_point_count(vb);++i)
        DEBUG_p("%d: (%d,%d)",i,(int)p[i].x,(int)p[i].y);

}

static void
_debug_segments(vector_buffer_t *vb)
{
    vector_segment_t *seg = vector_buffer_get_segment(vb,0);
    int i;
    for(i=0;i<vector_buffer_segment_count(vb);++i){
        point_t *p = vector_buffer_get_point(vb, seg[i].point);
        DEBUG_p("%d: type:%d point:%d:(%d,%d)",i,seg[i].type,seg[i].point,(int)p->x,(int)p->y);
    }

}


static void
_debug_shapes(vector_buffer_t *vb)
{
    shape_t *shape = vector_buffer_get_shape(vb,0);
    int i;
    for(i=0;i<vector_buffer_shape_count(vb);++i){
        DEBUG_p("segments: %d-%d",shape[i].path.segment,shape[i].path.segment+shape[i].path.segment_count-1);
    }
}


static void
_debug_anims(vector_buffer_t *vb){
    int i;
    for(i=0;i<vector_buffer_anim_count(vb);++i){
        anim_t *anim = vector_buffer_get_anim(vb,i);
        switch(anim->type){
        case AT_BALANCE:
            DEBUG_p("%d: balance: (shape %d): %d [%d %d]",i,anim->balance.shape,anim->balance.point,anim->balance.prev,anim->balance.next);
            break;
        case AT_CYCLE_COLORS:
            DEBUG_p("%d: cycle colors",i);
            break;
        case AT_MOVE_POINT:
            DEBUG_p("%d: move point: %d..%d (%d,%d)",i,anim->move_point.first, anim->move_point.last, anim->move_point.dx, anim->move_point.dy);
            break;
        case AT_MOVE_SHAPE:
            DEBUG_p("%d: move shape",i);
            break;
        case AT_PATH_SHAPE:
            DEBUG_p("%d: path shape",i);
            break;
        case AT_ROTATE:
            DEBUG_p("%d: rotate",i);
            break;
        case AT_SIZE:
            DEBUG_p("%d: size",i);
            break;
        }
    }
}


static void
_balance(const shape_t *shape, int nr, int prev, int next, const vector_segment_t *seg, point_t *point)
{
    int first;
    int last;
    _shape_points(shape, seg, &first, &last);
    prev += nr;
    if(prev < first)
        prev += shape->path.point_count;
    next += nr;
    if(next > last)
        next -= shape->path.point_count;
    point[nr].x = (point[prev].x+point[next].x)/2;
    point[nr].y = (point[prev].y+point[next].y)/2;
}


static void
_set_rotation(rotation_t *r, double rx, double ry, double rz){
    double cosa = cos(rz);
    double sina = sin(rz);
    double cosb = cos(ry);
    double sinb = sin(ry);
    double cosc = cos(rx);
    double sinc = sin(rx);
    r->xx = cosa*cosb;
    r->xy = cosa*sinb*sinc - sina*cosc;
    r->yx = sina*cosb;
    r->yy = sina*sinb*sinc + cosa*cosc;
    r->zx = -sinb;
    r->zy = cosb*sinc;
}


static void
_rotate_point(point_t *p, double cx, double cy, rotation_t *r){
    double x = p->x - cx;
    double y = p->y - cy;
    double pz = -(r->zx*x + r->zy*y)/512 + 1;
    p->x = (r->xx*x + r->xy*y) * pz + cx;
    p->y = (r->yx*x + r->yy*y) * pz + cy;
}


static void
_scale_point(point_t *p, double cx, double cy, double s)
{
    double x = p->x - cx;
    double y = p->y - cy;
    p->x = x*s + cx;
    p->y = y*s + cy;
}


static void
_move_shape(const shape_t *shape, double dx, double dy, const vector_segment_t *seg, point_t *point)
{
    int nr;
    int last;
    _shape_points(shape, seg, &nr, &last);
    for(;nr<=last;++nr){
        point[nr].x += dx;
        point[nr].y += dy;
    }
}


static void
_rotate_shape(const shape_t *shape, double cx, double cy, rotation_t *r, const vector_segment_t *seg, point_t *point)
{
    int nr;
    int last;
    _shape_points(shape, seg, &nr, &last);
    for(;nr<=last;++nr)
        _rotate_point(&point[nr], cx, cy, r);
}


static void
_scale_shape(const shape_t *shape, double cx, double cy, double scale, const vector_segment_t *seg, point_t *point)
{
    int nr;
    int last;
    _shape_points(shape, seg, &nr, &last);
    for(;nr<=last;++nr)
        _scale_point(&point[nr], cx, cy, scale);
}


#define add_if_above(x,limit,val) if(x>=(limit)) x+=(val)
#define sub_if_above(x,limit,val) if(x>=(limit)+(val)) x-=(val)


static point_t *
_insert_points(vector_buffer_t *vb, array_t *killpoint_array, int pos, int count)
{
    _debug_points(vb);
    point_t *p = vector_buffer_insert_points(vb, pos, count);
    vector_segment_t *seg = vector_buffer_get_segment(vb,0);
    shape_t *shape = vector_buffer_get_shape(vb,0);
    int i;

    for(i=0;i<vector_buffer_segment_count(vb);++i)
        add_if_above(seg[i].point, pos, count);
    for(i=0;i<vector_buffer_shape_count(vb);++i)
        switch(shape[i].type){
        case VECTOR_RECT:
            add_if_above(shape[i].rect.point, pos, count);
            break;
        case VECTOR_PATH:
            /* no point references */
            break;
        }
    for(i=0;i<array_length(killpoint_array);++i){
        killpoint_t *kp = (killpoint_t*)array_get(killpoint_array,i);
        add_if_above(kp->point, pos, count);
    }
    _debug_points(vb);
    return p;
}


static void
_insert_points_update_animations(vector_buffer_t *vb, int pos, int count)
{
    vector_segment_t *seg = vector_buffer_get_segment(vb,0);
    shape_t *shape = vector_buffer_get_shape(vb,0);
    int i;
    for(i=0;i<vector_buffer_anim_count(vb);++i){
        anim_t *anim = vector_buffer_get_anim(vb,i);
        switch(anim->type){
        case AT_BALANCE:
            add_if_above(anim->balance.point, pos, count);
            {
                int first;
                int last;
                _shape_points(&shape[anim->balance.shape], seg, &first, &last);
                if(pos >= first && pos <= last){
                    int pointnr;
                    pointnr = anim->balance.point + anim->balance.prev;
                    if(pointnr < first){
                        pointnr += shape[anim->balance.shape].path.point_count;
                        if(pos >= pointnr || pos < anim->balance.point)
                            --anim->balance.prev;
                    }else{
                        if(pos >= pointnr && pos < anim->balance.point)
                            --anim->balance.prev;
                    }
                    pointnr = anim->balance.point + anim->balance.next;
                    if(pointnr > last){
                        pointnr -= shape[anim->balance.shape].path.point_count;
                        if(pos > anim->balance.point || pos <= pointnr)
                            ++anim->balance.next;
                    }else{
                        if(pos > anim->balance.point && pos <= pointnr)
                            ++anim->balance.next;
                    }
                }
            }
            break;
        case AT_MOVE_POINT:
            add_if_above(anim->move_point.first, pos, count);
            add_if_above(anim->move_point.last, pos, count);
            break;
        case AT_PATH_SHAPE:
            add_if_above(anim->path_shape.point, pos, count);
            break;
        case AT_CYCLE_COLORS:
        case AT_MOVE_SHAPE:
        case AT_ROTATE:
        case AT_SIZE:
            /* no point references */
            break;
        }
    }
}


static void
_delete_points(vector_buffer_t *vb, array_t *killpoint_array, int pos, int count)
{
    vector_segment_t *seg = vector_buffer_get_segment(vb,0);
    shape_t *shape = vector_buffer_get_shape(vb,0);
    anim_t *anim = vector_buffer_get_anim(vb,0);
    int i;

    vector_buffer_delete_points(vb, pos, count);
    for(i=0;i<vector_buffer_segment_count(vb);++i)
        sub_if_above(seg[i].point, pos, count);
    for(i=0;i<vector_buffer_shape_count(vb);++i)
        switch(shape[i].type){
        case VECTOR_RECT:
            sub_if_above(shape[i].rect.point, pos, count);
            break;
        case VECTOR_PATH:
            /* no point references */
            break;
        }
    for(i=0;i<vector_buffer_anim_count(vb);++i)
        switch(anim[i].type){
        case AT_BALANCE:
            sub_if_above(anim[i].balance.point, pos, count);
            break;
        case AT_MOVE_POINT:
            sub_if_above(anim[i].move_point.first, pos, count);
            sub_if_above(anim[i].move_point.last, pos, count);
            break;
        case AT_PATH_SHAPE:
            sub_if_above(anim[i].path_shape.point, pos, count);
            break;
        case AT_CYCLE_COLORS:
        case AT_MOVE_SHAPE:
        case AT_ROTATE:
        case AT_SIZE:
            /* no point references */
            break;
        }
    for(i=0;i<array_length(killpoint_array);++i){
        killpoint_t *kp = (killpoint_t*)array_get(killpoint_array,i);
        sub_if_above(kp->point, pos, count);
    }
}


static vector_segment_t *
_insert_segment(vector_buffer_t *vb, int pos)
{
    _debug_segments(vb);
    vector_segment_t *seg = vector_buffer_insert_segment(vb, pos);
    int i;
    DEBUG_p("inserting segment at %d",pos);
    for(i=0;i<vector_buffer_shape_count(vb);++i){
        shape_t *shape = vector_buffer_get_shape(vb,i);
        switch(shape->type){
        case VECTOR_RECT:
            break;
        case VECTOR_PATH:
            add_if_above(shape->path.segment, pos, 1);
            break;
        }
    }
    _debug_segments(vb);
    return seg;
}


static void
_delete_segments(vector_buffer_t *vb, int pos, int count)
{
    int i;
    vector_buffer_delete_segments(vb, pos, count);
    for(i=0;i<vector_buffer_shape_count(vb);++i){
        shape_t *shape = vector_buffer_get_shape(vb,i);
        switch(shape->type){
        case VECTOR_RECT:
            break;
        case VECTOR_PATH:
            sub_if_above(shape->path.segment, pos, count);
            break;
        }
    }
}


/* <shape> <pos> */
static int
_handle_insert_point(vector_buffer_t *vb, array_t *killpoint_array, str_t *argv, int argc)
{
    int shapenr;
    int pointoffs;
    shape_t *shape;
    int segnr;
    int first;
    int last;
    int pointnr;
    vector_segment_t *seg;
    anim_t *anim;

    if(argc< 2)
        return -1;
    shapenr = _str_toi(&argv[0]);
    pointoffs = _str_toi(&argv[1]);
    if(shapenr >= vector_buffer_shape_count(vb))
        return -1;
    shape = vector_buffer_get_shape(vb,shapenr);
    /* Only path shapes can have points added to them. */
    if(shape->type != VECTOR_PATH)
        return -1;

    segnr = shape->path.segment;
    _shape_points(shape,vector_buffer_get_segment(vb,0),&first,&last);
    pointnr = first + pointoffs;
    if(pointnr >= vector_buffer_point_count(vb))
        return -1;
    DEBUG_p("segnr: %d, offs: %d",segnr,pointoffs);
    /* Get current points to calculate coordinates for the new one. */
    if(0==pointoffs){
        /* If the new point is the first one in a shape the previously
         * first segment has to be converted from type move to type
         * line.
         */
        seg = vector_buffer_get_segment(vb,segnr);
        seg->type = VECTOR_LINE;
    }
    /* The point do not need a value since it will be handled by the
     * balance animation.
     */
    _insert_points(vb, killpoint_array, pointnr, 1);

    seg = _insert_segment(vb,segnr+pointoffs);
    seg->type = 0==pointoffs ? VECTOR_GOTO : VECTOR_LINE;
    seg->point = pointnr;
    shape->path.segment = segnr;
    ++shape->path.segment_count;
    ++shape->path.point_count;

    /* The animations need to be updated after the shape parameters have
     * been updated since if depends on them to be correct.
     */
    _insert_points_update_animations(vb, pointnr, 1);

    anim = vector_buffer_add_anim(vb);
    anim->type = AT_BALANCE;
    anim->balance.shape = shapenr;
    anim->balance.point = pointnr;
    anim->balance.prev = -1;
    anim->balance.next = 1;

    //_debug_points(vb);
    _debug_segments(vb);
    _debug_shapes(vb);
    _debug_anims(vb);
    return 0;
}


/* <shape> <point> */
static int
_handle_kill_point(vector_buffer_t *vb, array_t *killpoint_array, int pointnr, int shapenr)
{
    shape_t *shape;
    int first;
    int last;
    int segnr;
    if(shapenr >= vector_buffer_shape_count(vb))
        return -1;
    shape = vector_buffer_get_shape(vb,shapenr);
    if(shape->type != VECTOR_PATH)
        return -1;

    _shape_points(shape,vector_buffer_get_segment(vb,0),&first,&last);
    if(pointnr < first || pointnr > last)
        return -1;
    segnr = shape->path.segment + pointnr - first;
    DEBUG_p("segnr: %d, pointnt: %d", segnr, pointnr);
    _delete_segments(vb, segnr, 1);
    _delete_points(vb, killpoint_array, pointnr, 1);
    if(pointnr == first){
        /* If the point is the first one in a shape the next segment
         * has to be converted from type line to type goto
         */
        vector_segment_t *seg = vector_buffer_get_segment(vb,segnr);
        seg->type = VECTOR_GOTO;
    }
    //shape->path.segment = segnr;
    --shape->path.segment_count;
    --shape->path.point_count;
    return 0;
}


/* <shape> <???> <color> <x> <y> <z?> <width> <height> */
static int
_handle_insert_cblock(vector_buffer_t *vb, str_t *argv, int argc)
{
    shape_t *shape;
    point_t *point;
    if(argc < 8)
        return -1;
    shape = vector_buffer_add_shape(vb);
    DEBUG_p("path number %d = %d",(int)vector_buffer_shape_index(vb,shape), _str_toi(&argv[0]));

    shape->type = VECTOR_RECT;
    shape->style = 0; /* _str_toi(&argv[1]); */
    shape->color = _str_toi(&argv[2]);
    shape->rect.point = vector_buffer_point_count(vb);
    point = vector_buffer_add_points(vb,1);
    point->x = _str_toi(&argv[3]);
    point->y = _str_toi(&argv[4]);
    shape->rect.width = _str_toi(&argv[6]);
    shape->rect.height = _str_toi(&argv[7]);
    return 0;
}


/* <shape> <style> <color> <points> <color?> <???> <???> [<point> ...] */
static int
_handle_insert_poly(vector_buffer_t *vb, str_t *argv, int argc)
{
    shape_t *shape;
    int count;
    if(argc < 8)
        return -1;
    count = _str_toi(&argv[3]);
    shape = vector_buffer_add_shape(vb);
    DEBUG_p("path number %d = %d",(int)vector_buffer_shape_index(vb,shape), _str_toi(&argv[0]));

    shape->type = VECTOR_PATH;
    shape->style = _str_toi(&argv[1]);
    shape->color = _str_toi(&argv[2]);
    shape->path.segment = vector_buffer_segment_count(vb);
    shape->path.segment_count = 0;
    shape->path.point_count = 0;
    if(argc >= 7+count){
        //point_t *point = vector_buffer_add_points(vb,count);
        int i;
        for(i=0;i<count;++i){
            point_t *p;
            if(i){
                p = vector_buffer_push_segment(vb, &shape->path, VECTOR_LINE);
            }else{
                p = vector_buffer_push_segment(vb, &shape->path, VECTOR_GOTO);
            }
            _read_point(p,&argv[7+i]);
        }
        if(VECTOR_OUTLINE == shape->style)
            vector_buffer_push_segment(vb, &shape->path, VECTOR_CLOSE);
    }
    return 0;
}


/* <x> <y> <z> <???> <???> <angle> <ry> <rx> <rz> */
static void
_handle_rotate_poly(vector_buffer_t *vb, int first, int last, str_t *argv){
    anim_t *anim = vector_buffer_add_anim(vb);
    anim->type = AT_ROTATE;
    anim->rotate.first = first;
    anim->rotate.last = last;
    anim->rotate.cx = _str_toi(&argv[0]);
    anim->rotate.cy = _str_toi(&argv[1]);
    anim->rotate.rx = (double)_str_toi(&argv[6])/16384;
    anim->rotate.ry = (double)_str_toi(&argv[7])/16384;
    anim->rotate.rz = (double)_str_toi(&argv[8])/16384;
    anim->rotate.angle = M_PI * 2 * -_str_toi(&argv[5])/1024;
    /* Ani ST uses hardcoded values in these cases. */
    if(anim->rotate.rx > 0.1)
        anim->rotate.cx = 160;
    if(anim->rotate.ry > 0.1)
        anim->rotate.cy = 100;
}


static void
_finish_keyframe(keyframe_t *kf, vector_buffer_t *vb, array_t *killpoint_array, color_t *color)
{
    point_t *point = vector_buffer_get_point(vb,0);
    vector_segment_t *seg = vector_buffer_get_segment(vb,0);
    shape_t *shape = vector_buffer_get_shape(vb,0);
    killpoint_t *killpoint = (killpoint_t*)array_get(killpoint_array, 0);
    int i;

    DEBUG_p("--- frame: %d ---",(int)array_index(&vb->keyframe_array,kf));
    _debug_shapes(vb);
    _debug_segments(vb);

    memcpy(kf->color, color, COLORS_MAX*sizeof(color_t));
    kf->point = vector_buffer_copy_points(vb);
    kf->point_count = vector_buffer_point_count(vb);
    kf->segment = vector_buffer_copy_segments(vb);
    kf->segment_count = vector_buffer_segment_count(vb);
    kf->shape = vector_buffer_copy_shapes(vb);
    kf->shape_count = vector_buffer_shape_count(vb);
    kf->anim = vector_buffer_copy_anims(vb);
    kf->anim_count = vector_buffer_anim_count(vb);

    /* apply all animations */
    for(i=0;i<vector_buffer_anim_count(vb);++i){
        const anim_t *anim = vector_buffer_get_anim(vb,i);
        int j;
        switch(anim->type){
        case AT_BALANCE:
            _balance(&shape[anim->balance.shape], anim->balance.point, anim->balance.prev, anim->balance.next, seg, point);
            break;
        case AT_CYCLE_COLORS:
            _memcycle(&color[anim->cycle_colors.first],
                      anim->cycle_colors.count*sizeof(color_t),
                      -anim->cycle_colors.direction*anim->cycle_colors.steps*sizeof(color_t));
            break;
        case AT_MOVE_POINT:
            for(j=anim->move_point.first;j<=anim->move_point.last;++j){
                point[j].x += anim->move_point.dx;
                point[j].y += anim->move_point.dy;
            }
            break;
        case AT_MOVE_SHAPE:
            for(j=anim->move_shape.first;j<=anim->move_shape.last;++j)
                _move_shape(&shape[j], anim->move_shape.dx, anim->move_shape.dy, seg, point);
            break;
        case AT_PATH_SHAPE:
            {
                point_t delta = point[anim->path_shape.point + anim->path_shape.steps];
                for(j=anim->path_shape.first;j<=anim->path_shape.last;++j)
                    _move_shape(&shape[j], delta.x, delta.y, seg, point);
            }
            break;
        case AT_ROTATE:
            {
                rotation_t r;
                _set_rotation(&r, anim->rotate.rx*anim->rotate.angle,
                                  anim->rotate.ry*anim->rotate.angle,
                                  anim->rotate.rz*anim->rotate.angle);
                for(j=anim->rotate.first;j<=anim->rotate.last;++j)
                    _rotate_shape(&shape[j], anim->rotate.cx, anim->rotate.cy, &r, seg, point);
                break;
            }
            break;
        case AT_SIZE:
            for(j=anim->size.first;j<=anim->size.last;++j)
                _scale_shape(&shape[j], anim->size.cx, anim->size.cy, anim->size.scale, seg, point);
            break;
        }
    }

    /* delete all animations */
    for(i=0;i<vector_buffer_anim_count(vb);++i){
        const anim_t *anim = vector_buffer_get_anim(vb,i);
        switch(anim->type){
        case AT_BALANCE:
        case AT_CYCLE_COLORS:
        case AT_MOVE_POINT:
        case AT_MOVE_SHAPE:
        case AT_ROTATE:
        case AT_SIZE:
            break;
        case AT_PATH_SHAPE:
            _delete_points(vb, killpoint_array, anim->path_shape.point, anim->path_shape.steps+1);
            break;
        }
    }
    vb->anim_array.members = 0;

    /* kill points */
    for(i=0;i<array_length(killpoint_array);++i)
        _handle_kill_point(vb, killpoint_array, killpoint[i].point, killpoint[i].shape);
    killpoint_array->members = 0;

    _debug_shapes(vb);
    _debug_segments(vb);
}


static int
_anist_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    vector_buffer_t vb;
    array_t killpoint_array;
    color_t color[COLORS_MAX];
    int version = 1;
    int speed = 32;
    double time = 0;
    keyframe_t *kf = NULL;
    str_t sn;

    vector_buffer_init(&vb, sizeof(shape_t), sizeof(keyframe_t), sizeof(anim_t));
    array_init(&killpoint_array, sizeof(killpoint_t));
    _set_default_palette(color);
    sn.str = data;
    sn.len = len;
    while(sn.len>0){
        command_t cmd;
        if(_read_command(&cmd,&sn)<0)
            break;
        if(!_str_cmp(&cmd.param[0],"act")){
            str_t *action;
            str_t *argv;
            int argc;
            if(cmd.param_count < 2)
                continue;
            /*
             * The syntax of the action statement varies between versions.
             *
             * variant 1: *act <command-num?> <command-string> args...
             * variant 2: *act <command-string> args...
             *
             */
            if(!isdigit(cmd.param[1].str[0])){
                action = &cmd.param[1];
                argv = &cmd.param[2];
                argc = cmd.param_count - 2;
            }else{
                if(cmd.param_count < 3)
                    continue;
                action = &cmd.param[2];
                argv = &cmd.param[3];
                argc = cmd.param_count - 3;
            }
            if(!_str_cmp(action,"CHANGE_COLOR")){
                /* <shape> <color> */
                if(argc>=2){
                    shape_t *shape = vector_buffer_get_shape(&vb,_str_toi(&argv[0]));
                    shape->color = _str_toi(&argv[1]);
                }
            }else if(!_str_cmp(action,"CHANGE_OCOLOR")){
                /* <shape?> <color?> */
                if(argc>=2){
                    shape_t *shape = vector_buffer_get_shape(&vb,_str_toi(&argv[0]));
                    shape->color = _str_toi(&argv[1]);
                }
            }else if(!_str_cmp(action,"CYCLE_COLORS")){
                /* <direction?> <first color> <last color> <steps> */
                if(argc >= 4){
                    anim_t *anim = vector_buffer_add_anim(&vb);
                    anim->type = AT_CYCLE_COLORS;
                    anim->cycle_colors.direction = _str_toi(&argv[0]);
                    anim->cycle_colors.first = _str_toi(&argv[1]);
                    anim->cycle_colors.count = _str_toi(&argv[2]);
                    anim->cycle_colors.steps = _str_toi(&argv[3]);
                }
            }else if(!_str_cmp(action,"INIT_COLORS")){
                /* <???> <first color> <count> */
                if(argc>=4){
                    int first = _str_toi(&argv[1]);
                    int count = _str_toi(&argv[2]);
                    if(first+count <= COLORS_MAX && argc >= 3+count){
                        int i;
                        for(i=0;i<count;++i)
                            _read_color(&color[first+i],&argv[3+i]);
                    }
                }
            }else if(!_str_cmp(action,"INSERT_CBLOCK")){
                _handle_insert_cblock(&vb, argv, argc);
            }else if(!_str_cmp(action,"INSERT_POINT")){
                _handle_insert_point(&vb, &killpoint_array, argv, argc);
            }else if(!_str_cmp(action,"INSERT_POLY")){
                //DEBUG_p("insert_poly (%.*s)",(int)cmd.param[1].len,cmd.param[1].str);
                _handle_insert_poly(&vb, argv, argc);
            }else if(!_str_cmp(action,"KILL_POINT")){
                /* <shape> <point> */
                if(argc >= 2){
                    int shapenr = _str_toi(&argv[0]);
                    if(shapenr < vector_buffer_shape_count(&vb)){
                        killpoint_t *kp = (killpoint_t*)array_append(&killpoint_array, 1);
                        shape_t *shape = vector_buffer_get_shape(&vb,shapenr);
                        int first;
                        int last;
                        _shape_points(shape, vector_buffer_get_segment(&vb,0), &first, &last);
                        kp->point = first + _str_toi(&argv[1]);
                        /* For some reason this value can be outside bounds
                         * for some files (created with Ani ST). Not sure
                         * what is the correct value to change it to, but
                         * at least this works with every file I have tried.
                         */
                        if(kp->point > last)
                            kp->point = last;
                        kp->shape = shapenr;
                    }
                }
            }else if(!_str_cmp(action,"KILL_POLY")){
                /* <shape> */
                if(argc >= 1){
                    int shapenr = _str_toi(&argv[0]);
                    if(shapenr < vector_buffer_shape_count(&vb)){
                        shape_t *shape = vector_buffer_get_shape(&vb,shapenr);
                        int first;
                        int last;
                        _shape_points(shape, vector_buffer_get_segment(&vb,0), &first, &last);
                        _delete_points(&vb, &killpoint_array, first, last - first +1);
                        switch(shape->type){
                        case VECTOR_RECT:
                            break;
                        case VECTOR_PATH:
                            _delete_segments(&vb,shape->path.segment,shape->path.segment_count);
                            break;
                        }
                        vector_buffer_delete_shape(&vb,shapenr);
                    }
                }
                _debug_shapes(&vb);
                _debug_segments(&vb);
            }else if(!_str_cmp(action,"MOVE_POINT")){
                /* <shape> <point> <dx> <dy> <dz?> */
                if(argc >= 4){
                    int shape = _str_toi(&argv[0]);
                    if(shape < vector_buffer_shape_count(&vb)){
                        int first;
                        int last;
                        _shape_points(vector_buffer_get_shape(&vb,shape),vector_buffer_get_segment(&vb,0),&first,&last);
                        int point = first + _str_toi(&argv[1]);
                        if(point <= last){
                            anim_t *anim = vector_buffer_add_anim(&vb);
                            anim->type = AT_MOVE_POINT;
                            anim->move_point.first = anim->move_point.last = point;
                            anim->move_point.dx = _str_toi(&argv[2]);
                            anim->move_point.dy = _str_toi(&argv[3]);
                        }
                    }
                }
            }else if(!_str_cmp(action,"MOVE_POLY")){
                /* <shape> <dx> <dy> <dz?> */
                if(argc >= 3){
                    int shape = _str_toi(&argv[0]);
                    if(shape < vector_buffer_shape_count(&vb)){
                        anim_t *anim = vector_buffer_add_anim(&vb);
                        anim->type = AT_MOVE_SHAPE;
                        anim->move_shape.first = shape;
                        anim->move_shape.last = shape;
                        anim->move_shape.dx = _str_toi(&argv[1]);
                        anim->move_shape.dy = _str_toi(&argv[2]);
                    }
                }
            }else if(!_str_cmp(action,"MOVE_POLYS")){
                /* <first shape> <shape count> <dx> <dy> <???> */
                if(argc >= 4){
                    int first = _str_toi(&argv[0]);
                    int last = first + _str_toi(&argv[1]) -1;
                    if(last < vector_buffer_shape_count(&vb)){
                        anim_t *anim = vector_buffer_add_anim(&vb);
                        anim->type = AT_MOVE_SHAPE;
                        anim->move_shape.first = first;
                        anim->move_shape.last = last;
                        anim->move_shape.dx = _str_toi(&argv[2]);
                        anim->move_shape.dy = _str_toi(&argv[3]);
                    }
                }
            }else if(!_str_cmp(action,"MOVE_SEGMENT")){
                /* <shape> <first point> <point count?> <???> <dx> <dy> <dz> */
                if(argc >= 6){
                    int shape = _str_toi(&argv[0]);
                    if(shape < vector_buffer_shape_count(&vb)){
                        int shape_first;
                        int shape_last;
                        int move_first;
                        int move_last;
                        _shape_points(vector_buffer_get_shape(&vb,shape),vector_buffer_get_segment(&vb,0),&shape_first,&shape_last);
                        move_first = _str_toi(&argv[1]) + shape_first;
                        move_last = move_first + _str_toi(&argv[2]) -1;
                        if(move_last <= shape_last){
                            anim_t *anim = vector_buffer_add_anim(&vb);
                            anim->type = AT_MOVE_POINT;
                            anim->move_point.first = move_first;
                            anim->move_point.last = move_last;
                            anim->move_point.dx = _str_toi(&argv[4]);
                            anim->move_point.dy = _str_toi(&argv[5]);
                        }
                    }
                }
            }else if(!_str_cmp(action,"PATH_POLY")){
                /* <shape> <steps> [<point> ...] */
                if(argc >= 3){
                    int steps = _str_toi(&argv[1]);
                    if(argc >= 2+steps){
                        anim_t *anim = vector_buffer_add_anim(&vb);
                        point_t *p;
                        int i;
                        anim->type = AT_PATH_SHAPE;
                        anim->path_shape.first = anim->path_shape.last = _str_toi(&argv[0]);
                        anim->path_shape.steps = steps -1;
                        anim->path_shape.point = vector_buffer_point_count(&vb);
                        p = vector_buffer_add_points(&vb,steps);
                        for(i=0;i<steps;++i)
                            _read_point(&p[i],&argv[2+i]);
                    }
                }
            }else if(!_str_cmp(action,"PATH_POLYS")){
                /* <first shape> <shape count> <steps> [<point> ...] */
                if(argc >= 4){
                    int steps = _str_toi(&argv[2]);
                    if(argc >= 3+steps){
                        int first = _str_toi(&argv[0]);
                        anim_t *anim = vector_buffer_add_anim(&vb);
                        point_t *p;
                        int i;
                        anim->type = AT_PATH_SHAPE;
                        anim->path_shape.first = first;
                        anim->path_shape.last = first + _str_toi(&argv[1]) -1;
                        anim->path_shape.steps = steps -1;
                        anim->path_shape.point = vector_buffer_point_count(&vb);
                        p = vector_buffer_add_points(&vb,steps);
                        for(i=0;i<steps;++i)
                            _read_point(&p[i],&argv[3+i]);
                    }
                }
            }else if(!_str_cmp(action,"REMOVE_BACKGROUND")){
                /* <???> */
                DEBUG_p("remove_background (%.*s)",(int)cmd.param[1].len,cmd.param[1].str);
                /* FIXME!! what should be done here? */
            }else if(!_str_cmp(action,"ROTATE_POLY")){
                /* <polygon> <x> <y> <z> <???> <???> <angle> <ry> <rx> <rz> */
                if(argc >= 10){
                    int shape = _str_toi(&argv[0]);
                    if(shape < vector_buffer_shape_count(&vb))
                        _handle_rotate_poly(&vb, shape, shape, &argv[1]);
                }
            }else if(!_str_cmp(action,"ROTATE_POLYS")){
                /* <first shape> <shape count> <cx> <cy> <z> <???> <???> <angle> <ry> <rx> <rz> */
                if(argc >= 11){
                    int first = _str_toi(&argv[0]);
                    int last = first + _str_toi(&argv[1]) -1;
                    if(last < vector_buffer_shape_count(&vb))
                        _handle_rotate_poly(&vb, first, last, &argv[2]);
                }
            }else if(!_str_cmp(action,"SIZE_POLY")){
                /* <shape> <cx> <cy> <???> <???> <???> */
                if(argc >= 6){
                    int shape = _str_toi(&argv[0]);
                    if(shape < vector_buffer_shape_count(&vb)){
                        anim_t *anim = vector_buffer_add_anim(&vb);
                        anim->type = AT_SIZE;
                        anim->size.first = shape;
                        anim->size.last = shape;
                        anim->size.cx = _str_toi(&argv[1]);
                        anim->size.cy = _str_toi(&argv[2]);
                        anim->size.scale = (double)_str_toi(&argv[4])/_str_toi(&argv[5]);
                    }
                }
            }else if(!_str_cmp(action,"TWEEN_COLOR")){
                /* <direction?> <color> <red delta?> <green delta?> <blue delta?> */
                DEBUG_p("animation of type tween color not supported");
                /* FIXME!! */
            }else{
                DEBUG_p("unhandled action: '%.*s'",(int)action->len,action->str);
            }
        }else if(!_str_cmp(&cmd.param[0],"tween")){
            /* *tween <start time> <duration> <end time> <actions?> */
            if(kf)
                _finish_keyframe(kf, &vb, &killpoint_array, color);
            DEBUG_p("---inserting keyframe");
            kf = vector_buffer_begin_keyframe(&vb);
            if(cmd.param_count >= 3){
                kf->start = time;
                /* This calculation follows the specification in the Ani ST
                 * manual. It says 32 is normal speed and a "global speed"
                 * of 64 would cause the animation to play twice as fast.
                 * It also says the time is measured int jiffies (1/60 of a
                 * second), but the time written is 10/3 times larger, in
                 * other words the unit in the file is 1/200 fd a second.
                 * However Ani ST itself is actually rendering at a slightly
                 * slower speed most of the time. So while this could be
                 * considered technically correct, it might also be faster
                 * than originally intended.
                 */
                kf->duration = (double)(32*_str_toi(&cmd.param[2])/speed)/200;
            }
            time += kf->duration;
        }else if(!_str_cmp(&cmd.param[0],"script")){
            /* *script <filename> <???> <width> <height> */
            if(cmd.param_count>=5){
                h->info->width = _str_toi(&cmd.param[3]);
                h->info->height = _str_toi(&cmd.param[4]);
            }
        }else if(!_str_cmp(&cmd.param[0],"version")){
            /* *version <version> */
            if(cmd.param_count>=2)
                version = _str_toi(&cmd.param[1]);
            DEBUG_p("version = %d",version);
        }else if(!_str_cmp(&cmd.param[0],"speed")){
            /* *speed <speed> */
            if(cmd.param_count>=2)
                speed = MAX(1,_str_toi(&cmd.param[1]));
        }else{
            DEBUG_p("unhandled command: '%.*s'",(int)cmd.param[0].len,cmd.param[0].str);
        }
    }
    if(kf)
        _finish_keyframe(kf, &vb, &killpoint_array, color);
    h->info->frame_count = vector_buffer_keyframe_count(&vb);
    h->keyframe = vector_buffer_copy_keyframes(&vb);
    vector_buffer_done(&vb);
    array_done(&killpoint_array);
    return 0;
}


static void
_anist_render(abydos_plugin_handle_t *h,
              cairo_t *cr,
              int page,
              cairo_bool_t *layer_visibility,
              int variant,
              double time,
              int frame)
{
    keyframe_t *kf = &h->keyframe[frame];
    color_t color[COLORS_MAX];
    point_t *point;
    double progress = (time - kf->start) / kf->duration;
    int i;
    cairo_save(cr);
    cairo_rectangle(cr,0.005,0.005,h->info->width-0.005,h->info->height-0.005);
    cairo_clip(cr);

    memcpy(color,kf->color,COLORS_MAX*sizeof(color_t));
    point = (point_t*)malloc(kf->point_count*sizeof(point_t));
    memcpy(point, kf->point, kf->point_count*sizeof(point_t));

    cairo_save(cr);
    cairo_set_fill_rule (cr, CAIRO_FILL_RULE_EVEN_ODD);
    for(i=0;i<kf->anim_count;++i){
        int j;
        const anim_t *anim = &kf->anim[i];
        switch(anim->type){
        case AT_BALANCE:
            _balance(&kf->shape[anim->balance.shape], anim->balance.point, anim->balance.prev, anim->balance.next, kf->segment, point);
            break;
        case AT_CYCLE_COLORS:
            {
                double stepsize = 1.0/anim->cycle_colors.steps;
                int step = anim->cycle_colors.steps * progress;
                color_t tmp_color[COLORS_MAX];
                double s1 = progress/stepsize - step;
                double s2 = 1.0 - s1;
                for(j=0;j<anim->cycle_colors.count;++j){
                    int start_color = anim->cycle_colors.first+(anim->cycle_colors.count+j+(step)*anim->cycle_colors.direction)%anim->cycle_colors.count;
                    int end_color = anim->cycle_colors.first+(anim->cycle_colors.count+j+(step+1)*anim->cycle_colors.direction)%anim->cycle_colors.count;
                    color_t *c1 = &color[start_color];
                    color_t *c2 = &color[end_color];
                    tmp_color[j].r = c1->r*s2 + c2->r*s1;
                    tmp_color[j].g = c1->g*s2 + c2->g*s1;
                    tmp_color[j].b = c1->b*s2 + c2->b*s1;
                }
                memcpy(&color[anim->cycle_colors.first],&tmp_color,anim->cycle_colors.count*sizeof(color_t));
            }
            break;
        case AT_MOVE_POINT:
            {
                double dx = (double)anim->move_point.dx*progress;
                double dy = (double)anim->move_point.dy*progress;
                for(j=anim->move_point.first;j<=anim->move_point.last;++j){
                    point[j].x += dx;
                    point[j].y += dy;
                }
            }
            break;
        case AT_MOVE_SHAPE:
            {
                double dx = (double)anim->move_shape.dx * progress;
                double dy = (double)anim->move_shape.dy * progress;
                for(j=anim->move_shape.first;j<=anim->move_shape.last;++j)
                    _move_shape(&kf->shape[j], dx, dy, kf->segment, point);
            }
            break;
        case AT_PATH_SHAPE:
            {
                double stepsize = 1.0/anim->path_shape.steps;
                int step = anim->path_shape.steps * progress;
                double s1 = progress/stepsize - step;
                double s2 = 1.0 - s1;
                point_t *p = &point[anim->path_shape.point+step];
                double dx = p[0].x*s2 + p[1].x*s1;
                double dy = p[0].y*s2 + p[1].y*s1;
                for(j=anim->move_shape.first;j<=anim->move_shape.last;++j)
                    _move_shape(&kf->shape[j], dx, dy, kf->segment, point);
            }
            break;
        case AT_ROTATE:
            {
                rotation_t r;
                double a = anim->rotate.angle*progress;
                _set_rotation(&r, anim->rotate.rx*a, anim->rotate.ry*a, anim->rotate.rz*a);
                for(j=anim->rotate.first;j<=anim->rotate.last;++j)
                    _rotate_shape(&kf->shape[j], anim->rotate.cx, anim->rotate.cy, &r, kf->segment, point);
            }
            break;
        case AT_SIZE:
            {
                double scale = anim->size.scale*progress + (1.0-progress);
                for(j=anim->size.first;j<=anim->size.last;++j)
                    _scale_shape(&kf->shape[j], anim->size.cx, anim->size.cy, scale, kf->segment, point);
            }
            break;
        }
    }
    cairo_restore(cr);

    for(i=0;i<kf->shape_count;++i){
        const shape_t *shape = &kf->shape[i];
        color_t *c = &color[shape->color];
        int j;
        cairo_save(cr);
        cairo_set_source_rgb(cr,c->r,c->g,c->b);
        switch(shape->type){
        case VECTOR_RECT:
            cairo_rectangle(cr,point[shape->rect.point].x,point[shape->rect.point].y,shape->rect.width,shape->rect.height);
            break;
        case VECTOR_PATH:
            DEBUG_p("path with %d segments",shape->path.segment_count);
            for(j=0;j<shape->path.segment_count;++j){
                const vector_segment_t *seg = &kf->segment[shape->path.segment+j];
                switch(seg->type){
                case VECTOR_GOTO:
                    cairo_move_to(cr,point[seg->point].x,point[seg->point].y);
                    break;
                case VECTOR_LINE:
                    cairo_line_to(cr,point[seg->point].x,point[seg->point].y);
                    break;
                case VECTOR_CURVE:
                    /* not supported by Ani ST */
                    break;
                case VECTOR_CLOSE:
                    cairo_close_path(cr);
                    break;
                }
            }
            break;
        }
        if(VECTOR_FILLED == shape->style){
            cairo_fill(cr);
        }else{
            cairo_stroke(cr);
        }
        cairo_restore(cr);
    }
    cairo_restore(cr);
    free(point);
}


static void
_anist_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->keyframe[i].duration;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _anist_get_mime_types,
        .new = _anist_new,
        .free = _anist_free,
        .create_from_data = _anist_create_from_data,
        .render = _anist_render,
        .get_duration = _anist_get_duration
    };
    return &p;
}
