/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

/*

Header:

-------v------v--------------
offset | size | Description
-------+------+--------------
     0 |    3 | ID = "GX2"
     3 |    1 | version = 1
     4 |    2 | header size
     6 |    1 | bits per pixel
     7 |    2 | width
     9 |    2 | height
    11 |    2 | horizontal size
    13 |    2 | vertical size
    15 |    1 | unit

*/

typedef struct{
    const uint8_t *p;
    const uint8_t *end;
    int bytes;
    int runlength;
}rle_t;

static void
rle_init(rle_t *rle, const uint8_t *data, size_t len)
{
    rle->p = data;
    rle->end = data+len;
    rle->bytes = 0;
}

static int
rle_get(rle_t *rle)
{
    if(rle->bytes <= 0){
        if(rle->p >= rle->end)
            return -1;
        rle->bytes = *rle->p++;
        if(0==rle->bytes)
            DEBUG_p("bytes should not be zero");
        if(rle->bytes >= 128){
            if(rle->p >= rle->end){
                DEBUG_p("unexcpected end of data");
                return -1;
            }
            rle->runlength = *rle->p++;
            rle->bytes &= 127;
            if(rle->bytes < 4)
                DEBUG_p("too few bytes");
        }else{
            rle->runlength = -1;
        }
    }
    --rle->bytes;
    if(rle->runlength >= 0){
        return rle->runlength;
    }else{
        if(rle->p >= rle->end){
            DEBUG_p("unexcpected end of data");
            return -1;
        }
        return *rle->p++;
    }
}


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    enum{
        T_GX1,
        T_GX2
    }type;
    cairo_surface_t *surface;
};


static const char **
_gx2_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-gx1",
        "image/x-gx2",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_gx2_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    if(!strcasecmp(mime_type,"image/x-gx1")){
        h->type = T_GX1;
    }else{
        h->type = T_GX2;
    }
    h->surface = NULL;
    h->info->threadsafe = 1;
    return h;
}


static void
_gx2_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


#define get16(a) le16toh(*(uint16_t*)(a))
#define get32(a) le32toh(*(uint32_t*)(a))
//#define plane_stride(w) ((w-1)/8+1)
//#define map_stride(w) plane_stride(plane_stride(w))


static ssize_t
_handle_palette_2bit(uint32_t *dst, int depth, const uint8_t *data, size_t len)
{
    int colors = 1<<depth;
    const uint8_t *p = data;
    int i;
    if(len<colors)
        return -1;
    for(i=0;i<colors;++i){
        int r = ((*p>>1) & 2) | ((*p>>5) & 1);
        int g = ((*p>>0) & 2) | ((*p>>4) & 1);
        int b = ((*p<<1) & 2) | ((*p>>1) & 1);
        ++p;
        r |= r<<2 | r<<4 | r<<6;
        g |= g<<2 | g<<4 | g<<6;
        b |= b<<2 | b<<4 | b<<6;
        *dst++ = r<<16|g<<8|b<<0;
    }
    return colors;
}


static ssize_t
_handle_palette_8bit(uint32_t *dst, int depth, int type, const uint8_t *data, size_t len)
{
    int i;
    switch(type & 0xf){
    case 0:
        DEBUG_p("monochrome text");
        break;
    case 1:
        DEBUG_p("RGB text bitmap");
        break;
    case 2:
        DEBUG_p("RGB lookup table");
        {
            const uint8_t *p = data;
            int colors = 1<<depth;
            if(len<colors*3)
                return -1;
            for(i=0;i<colors;++i){
                int r = *p++;
                int g = *p++;
                int b = *p++;
                *dst++ = r<<16|g<<8|b<<0;
            }
            return p-data;
        }
    case 3:
        DEBUG_p("RGB pixels");
        break;
    }
    if(data[0] & 128){
        DEBUG_p("bit planes");
    }else{
        DEBUG_p("uniplanar");
    }
    DEBUG_p("block count: %d",data[1]);
    return -1;
}


/*
static int
_unpack_bitplane(uint8_t *dst, int width, int height, rle_t *rle)
{
}
*/


static uint8_t **
_bitplanes_new(int stride, int height, int depth)
{
    uint8_t **bitplane;
    int y;
    bitplane = (uint8_t**)malloc(depth*sizeof(uint8_t*));
    bitplane[0] = (uint8_t*)malloc(stride*height*depth*sizeof(uint8_t));
    for(y=1;y<depth;++y)
        bitplane[y] = bitplane[y-1] + stride*height;
    return bitplane;
}


static void
_bitplanes_free(uint8_t **bitplane)
{
    free(bitplane[0]);
    free(bitplane);
}


static int
_unpack_planes(uint8_t **bitplane, int stride, int height, int depth, const uint8_t *data, size_t len)
{

    uint8_t changemap[4096];
    int maplen = (stride+7)/8;
    rle_t rle;
    int plane;

    rle_init(&rle,data,len);

    for(plane=0;plane<depth;++plane){
        uint8_t *dst = bitplane[plane];
        int x;
        int y;
        int b;

        DEBUG_p("bytes @1: %d",rle.bytes);
        for(x=0;x<stride;++x){
            b = rle_get(&rle);
            if(b<0)
                return -1;
            *dst++ = b;
        }

        DEBUG_p("bytes @2: %d",rle.bytes);
        for(y=1;y<height;++y){
            for(x=0;x<maplen;++x){
                b = rle_get(&rle);
                if(b<0)
                    return -1;
                changemap[x] = b;
                //DEBUG_p("changemap %d = %x",x,b);
            }
            for(x=0;x<stride;++x){
                int bits;
                if(!(x&7)){
                    bits = changemap[x/8];
                }
                bits<<=1;
                if(bits&0x100){
                    //DEBUG_p("there is a change at %dx%d",x,y);
                    b = rle_get(&rle);
                    if(b<0)
                        return -1;
                    *dst++ = b;
                }else{
                    *dst = dst[-stride];
                    ++dst;
                }
            }
        }
    }
    DEBUG_p("bytes left: %d",(int)(rle.end-rle.p));
    return 0;
}


#if 0
static uint8_t **
_get_bitplanes(int width, int height, int depth, const uint8_t *data, size_t len)
{

    uint8_t changemap[4096];
    int stride = (width+7)/8;
    int maplen = (stride+7)/8;
    rle_t rle;
    uint8_t **bitplane;
    uint8_t *dst;
    int x;
    int y;
    int b;

    rle_init(&rle,data,len);
    bitplane = _bitplanes_new(stride,height,depth);
    dst = bitplane[0];

    DEBUG_p("bytes @1: %d",rle.bytes);
    for(x=0;x<stride;++x){
        b = rle_get(&rle);
        if(b<0)
            goto error;
        *dst++ = b;
    }

    DEBUG_p("bytes @2: %d",rle.bytes);
    for(y=1;y<height*depth;++y){
        for(x=0;x<maplen;++x){
            b = rle_get(&rle);
            if(b<0)
                goto error;
            changemap[x] = b;
            //DEBUG_p("changemap %d = %x",x,b);
        }
        for(x=0;x<stride;++x){
            int bits;
            if(!(x&7)){
                bits = changemap[x/8];
            }
            bits<<=1;
            if(bits&0x100){
                //DEBUG_p("there is a change at %dx%d",x,y);
                b = rle_get(&rle);
                if(b<0)
                    goto error;
                *dst++ = b;
            }else{
                *dst = dst[-stride];
                ++dst;
            }
        }
    }
    DEBUG_p("bytes left: %d",(int)(rle.end-rle.p));

    return bitplane;
error:
    _bitplanes_free(bitplane);
    return NULL;
}
#endif


static cairo_surface_t *
_surface_from_bitplanes(uint8_t **bitplane, int width, int height, int depth, const uint32_t *palette)
{
    int src_offs = 0;
    int src_stride = (width+7)/8;
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);

    while(height-->0){
        int x;
        for(x=0;x<src_stride;++x){
            int plane;
            uint8_t b[8];
            int i;
            memset(b,0,8);
            for(plane=0;plane<depth;++plane){
                for(i=0;i<8;++i)
                    if(bitplane[plane][src_offs+x] & (128>>i))
                        b[i] |= 1<<plane;
            }
            for(i=0;i<8;++i){
                if(x*8+i >= width)
                    break;
                dst[x*8+i] = palette[b[i]];
            }
        }
        src_offs += src_stride;
        dst += dst_stride;
    }

    cairo_surface_mark_dirty(surface);
    return surface;
}


static cairo_surface_t *
_surface_from_planar(uint8_t *src, int width, int height, const uint32_t *palette)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t) - width;
    while(height-->0){
        int x = width;
        while(x-->0)
            *dst++ = palette[*src++];
        dst += skip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


#if 0
static int
_try_unpack(const uint8_t *data, size_t len)
{
    rle_t rle;
    int count = 0;

    rle_init(&rle,data,len);

    for(;;){
        if(rle.bytes <= 0){
            if(rle.p >= rle.end)
                return count;
            rle.bytes = *rle.p++;
            if(0==rle.bytes){
                DEBUG_p("bytes should not be zero @ %d",count);
                //return -1;
            }
            if(rle.bytes >= 128){
                if(rle.p >= rle.end){
                    DEBUG_p("unexcpected end of data");
                    return -1;
                }
                rle.runlength = *rle.p++;
                rle.bytes &= 127;
                if(rle.bytes < 4){
                    DEBUG_p("too few bytes (illigal data)");
                    return -1;
                }
            }else{
                rle.runlength = -1;
            }
        }
        --rle.bytes;
        if(rle.runlength >= 0){
        }else{
            if(rle.p >= rle.end){
                DEBUG_p("unexcpected end of data");
                return -1;
            }
            ++rle.p;
        }
        ++count;
    }
}


static void
_analyze(int base, const uint8_t *data, size_t len)
{
    int cut;
    for(cut=0;cut<len;++cut){
        int offset;
        for(offset=base;offset<32 /* len-cut */;++offset){
            int result;
            DEBUG_p(">>> trying offset %d cut %d",offset,cut);
            result = _try_unpack(data+offset,len-offset-cut);
            if(result>=0){
                DEBUG_p("<<< SUCCESS: got %d bytes, offset: %d, cut: %d",result,offset,cut);
                return;
            }else{
                DEBUG_p("<<< failed");
            }
        }
    }
    /*
        offset 561, cut 0
        offset 2, cut 25
    */
}
#endif


static int
_handle_gx1(abydos_plugin_handle_t *h, const uint8_t *data, size_t len)
{
    uint32_t palette[256];
    uint8_t **bitplane;
    int size;
    int stride;
    int depth;
    int ret;

    if(data[0] != 0xfa){
        h->info->error = "Invalid data";
        return -1;
    }

#if 0
    _analyze(5,data,len);
    return -1;
#endif

    switch(data[1]){
    case 4:
        /* Header size = 3 */
        size = 3;
        data += size;
        len -= size;

        h->info->width = 640;
        h->info->height = 200;
        h->info->pixel_ratio = 5.0/12;
        depth = 1;

        palette[0] = 0;
        palette[1] = 0xffffff;
        break;
    case 13:
        /* Header size = 4 */
        size = 4;
        data += size;
        len -= size;

        h->info->width = 640;
        h->info->height = 350;
        h->info->pixel_ratio = 35.0/48;
        depth = 4;

        size = _handle_palette_2bit(palette, depth, data, len);
        if(size<0)
            return -1;
        data += size;
        len -= size;
        break;
    default:
        h->info->error = "Unknown file type";
        return -1;
    }

    stride = (h->info->width+7)/8;
    bitplane = _bitplanes_new(stride, h->info->height, depth);
    ret = _unpack_planes(bitplane, stride, h->info->height, depth, data, len);
    if(ret>=0)
        h->surface = _surface_from_bitplanes(bitplane, h->info->width, h->info->height, depth, palette);
    _bitplanes_free(bitplane);
    return ret;
}


static int
_handle_gx2(abydos_plugin_handle_t *h, const uint8_t *data, size_t len)
{
    int palette_type;
    uint32_t palette[256];
    ssize_t size;
    int depth;
    int screen_width;
    int screen_height;
    int ret;

    if(len<18){
        h->info->error = "Data too short";
        return -1;
    }
    if(memcmp(data,"GX2",3)){
        h->info->error = "Invalid data";
        return -1;
    }
    if(data[3] != 1){
        h->info->error = "Unknown version";
        return -1;
    }
    size = get16(data+4);
    DEBUG_p("header size: %d",(int)size);
    depth = data[6];
    DEBUG_p("bits per pixel: %d",depth);
    h->info->width = get16(data+7);
    h->info->height = get16(data+9);
    DEBUG_p("dimensions: %dx%d",h->info->width,h->info->height);
    screen_width = get16(data+11);
    screen_height = get16(data+13);
    DEBUG_p("horizontal size: %d",screen_width);
    DEBUG_p("vertical size: %d",screen_height);
    h->info->pixel_ratio = (double)(h->info->height*screen_width)/(h->info->width*screen_height);
    data += size;
    len -= size;

    palette_type = data[0];
    DEBUG_p("palette type: %d",palette_type);
    DEBUG_p("blocks: %d",data[1]);
    data += 2;
    len -= 2;

    size = _handle_palette_8bit(palette,depth,palette_type,(uint8_t*)data,len);
    if(size<0)
        return -1;
    data += size;
    len -= size;

    if(8==depth){
        uint8_t *pixels = (uint8_t*)malloc(h->info->width * h->info->height);
        ret = _unpack_planes(&pixels, h->info->width, h->info->height, 1, data, len);
        if(ret>=0)
            h->surface = _surface_from_planar(pixels, h->info->width, h->info->height, palette);
        free(pixels);
    }else{
        int stride = (h->info->width+7)/8;
        uint8_t **bitplane = _bitplanes_new(stride, h->info->height, depth);
        /* Decompress all bitplanes as one */
        ret = _unpack_planes(bitplane, stride, h->info->height * depth, 1, data, len);
        if(ret>=0)
            h->surface = _surface_from_bitplanes(bitplane, h->info->width, h->info->height, depth, palette);
        _bitplanes_free(bitplane);
    }
    return ret;
}


static int
_gx2_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    switch(h->type){
    case T_GX1:
        return _handle_gx1(h, (uint8_t*)data, len);
    case T_GX2:
        return _handle_gx2(h, (uint8_t*)data, len);
    }
    /* not reached */
    return -1;
}


static cairo_surface_t *
_gx2_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _gx2_get_mime_types,
        .new = _gx2_new,
        .free = _gx2_free,
        .create_from_data = _gx2_create_from_data,
        .get_image_surface = _gx2_get_image_surface,
    };
    return &p;
}
