/*
 * Copyright (C) 2018 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <autoconf.h>
#include <assert.h>
#include <ctype.h>                      /* isspace() */
#include <fcntl.h>                      /* O_NONBLOCK */
#include <limits.h>                     /* LONG_MAX */
#include <netinet/in.h>                 /* ntohs() */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/signal.h>
#include <sys/wait.h>
#include <unistd.h>                     /* dup2() */

#include <abydos-plugin.h>
#include <netpbm/pam.h>

#include "buffer.h"
#include "debug.h"

//#define ENABLE_ILBM
//#define ENABLE_JPEG
//#define ENABLE_JPEG2000
//#define ENABLE_NEOCHROME
//#define ENABLE_PCX
//#define ENABLE_SPECTRUM512

/*

Formats supported by netpbm:

  411     | Sony Mavic camera
  ascii   | ASCII
  atk     | Andrew Toolkit raster object
  avs     | AVS X image
  biorad  | Biorad confocal
  bmp     | Windows BMP or OS/2 BMP
  brush   | Doodle brush
  camera  | Camera raw image
  cis     | Compuserve RLE image
  cmuwm   | CMU window manager bitmap
  ddbug   | Diddle or DiddleBug sketch
  escp2   | ESC/P2 printer file
  eyuv    | Berkley YUV
  fits    | FITS
  fs      | Usenix FaceSaver
  g3      | Group 3 fax file
  gem     | GEM .img
  gif     | Compuserve GIF
  gould   | Gould scanner file
  hdiff   | Horizontal difference image
  hips    | Hunan Informational Processing Library HIPS
  hpcd    | Photo-CD image
  ilbm    | Interlaced Bitmap
  img     | Img-whatnot
  info    | Amiga .info icon
  jpek2k
  jpeg
  leaf    | Interleaf image
  lispm   | Lisp Machine bitmap
  macp    | MacPaint
  mda     | Microdesign .mda or .mdp
  mgr     | MGR window manager bitmap
  mtv     | MTV or PRT ray tracer output
  neo     | Neochrome
  palm    | Palm bitmap
  pc1
  pcdov   | Photo CD overview
  pcx
  pdbimg  | Palm Pilot PDB image
  pfm     | Portable Float Map
  pi1
  pi3
  pict    | Macintosh PICT
  pj      | HP PaintJet
  pk      | packed format font
  png
  psid    | PostScript image data
  ps
  qrt     | QRT raytracer output
  rast    | Sun resterfile
  rla     | Alias/Wavefront RLA and RPF image
  sbig    | Santa Barbera Instrument Group astronomical CCD camera file (CCDOPS)
  sgi     | SGI image
  sir     | Solitair Image Recorder file
  sld     | AutoCAD slide
  spc     | compressed spectrum file
  spot    | SPOT satellite image
  spu     | uncompressed spectrum file
  srf     | SRF image
  st4     | Santa Barbera Instrument Group astronimical ST-4 camera file
  sunicon | Sun icon
  svg
  tga
  thinkjet| HP ThinkJet printer file
  tiff
  wbmp    | wireless bitmap
  winicon | Windows icon
  xbm     | X10 or X11 bitmap
  xpm     | X11 pixmap
  xvmini  | XV thumbnail
  xwd     | X10 or X11 window dump
  ybm     | Bennet Yee face file
  zeiss   | Zeiss confocal file
*/


#define ATARI_PIXEL_RATIO (5.0/6)
//#define BUFFER_SIZE 4096


#ifndef MIN
# define MIN(a,b) ((a)<(b)?(a):(b))
#endif


#if 0
typedef enum {
    PNM_OK = 0,
    PNM_CONTINUE,
    PNM_ERROR
} result_t;

typedef enum {
    PNM_FORMAT_PGM = 1,
    PNM_FORMAT_PGM_RAW,
    PNM_FORMAT_PPM,
    PNM_FORMAT_PPM_RAW,
    PNM_FORMAT_PBM,
    PNM_FORMAT_PBM_RAW
} pnm_format_t;
#endif

//typedef void(*draw_func_t)(abydos_plugin_handle_t *,int pixels,int bytes);

struct _abydos_plugin_handle_t{
    cairo_surface_t *surface;
    uint8_t *pixels;        /* incoming pixel data buffer */
    int rowstride;
    abydos_plugin_info_t *info;

    /* external converter */
    const char *converter;
    int input[2];
    int output[2];
    int pid;

    /* progressive loading */
    abydos_buffer_t inbuf;
    abydos_plugin_info_func_t info_func;
    abydos_plugin_progress_func_t progress_func;
    void *userdata;
    unsigned int scan_state;
};



static const char **
_netpbm_get_mime_types()
{
    static const char *mime_types[] = {
#ifdef ENABLE_BMP
        "image/bmp",
#endif
#ifdef ENABLE_ILBM
        "image/x-ilbm",
#endif
#ifdef ENABLE_JPEG2000
        "image/jp2",
        "image/jpeg2000",
        "image/jpx",
#endif
#ifdef ENABLE_NEOCHROME
        "image/x-neo",
        "image/x-neochrome",
#endif
#ifdef ENABLE_PC1
        "image/x-pc1",
#endif
#ifdef ENABLE_PC2
        "image/x-pc2",
#endif
#ifdef ENABLE_PC3
        "image/x-pc3",
#endif
#ifdef ENABLE_PCX
        "image/vnd.zbrush.pcx",
        "image/x-pcx",
#endif
        "image/x-portable-anymap",
        "image/x-portable-arbitrarymap",
        "image/x-portable-bitmap",
        "image/x-portable-graymap",
        "image/x-portable-greymap",
        "image/x-portable-pixmap",
#ifdef ENABLE_SPECTRUM512
        "image/x-spectrum512-compressed",
        "image/x-spectrum512-uncompressed",
#endif
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_netpbm_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->pixel_ratio = 1;
#ifdef ENABLE_BMP
    if(!strcasecmp(mime_type,"image/bmp")){
        h->converter = "bmptopnm";
    }else
#endif
#ifdef ENABLE_ILBM
    if(!strcasecmp(mime_type,"image/x-ilbm")){
        h->converter = "ilbmtoppm";
    }else
#endif
#ifdef ENABLE_JPEG2000
    if(!strcasecmp(mime_type,"image/jp2")
            |!strcasecmp(mime_type,"image/jpeg2000")
            |!strcasecmp(mime_type,"image/jpx"))
    {
        h->converter = "jpeg2ktopam";
    }else
#endif
#ifdef ENABLE_NEOCHROME
    if(!strcasecmp(mime_type,"image/x-neo")
            |!strcasecmp(mime_type,"image/x-neochrome"))
    {
        h->converter = "neotoppm";
        h->info->pixel_ratio = ATARI_PIXEL_RATIO;
    }else
#endif
#ifdef ENABLE_PC1
    if(!strcasecmp(mime_type,"image/x-pc1")){
        h->converter = "pc1toppm";
        h->info->pixel_ratio = ATARI_PIXEL_RATIO;
    }else
#endif
#ifdef ENABLE_PI1
    if(!strcasecmp(mime_type,"image/x-pi1")){
        h->converter = "pi1toppm";
        h->info->pixel_ratio = ATARI_PIXEL_RATIO;
    }else
#endif
#ifdef ENABLE_PCX
    if(!strcasecmp(mime_type,"image/vnd.zbrush.pcx")
      |!strcasecmp(mime_type,"image/x-pcx"))
    {
        h->converter = "pcxtoppm";
    }else
#endif
    if(!strcasecmp(mime_type,"image/x-portable-anymap")|
       !strcasecmp(mime_type,"image/x-portable-arbitrarymap")|
       !strcasecmp(mime_type,"image/x-portable-bitmap")|
       !strcasecmp(mime_type,"image/x-portable-graymap")|
       !strcasecmp(mime_type,"image/x-portable-greymap")|
       !strcasecmp(mime_type,"image/x-portable-pixmap"))
    {
        h->converter = NULL;
    }else
#ifdef ENABLE_SPECTRUM512
    if(!strcasecmp(mime_type,"image/x-spectrum512-compressed")){
        h->converter = "spctoppm";
        h->info->pixel_ratio = ATARI_PIXEL_RATIO;
    }else if(!strcasecmp(mime_type,"image/x-spectrum512-uncompressed")){
        h->converter = "sputoppm";
        h->info->pixel_ratio = ATARI_PIXEL_RATIO;
    }else
#endif
    {
        free(h);
        return NULL;
    }
    info->threadsafe = 1;
    h->surface = NULL;
    return h;
}


static void
_netpbm_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static cairo_surface_t *
_surface_from_pam(abydos_plugin_handle_t *h, struct pam *pam)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,h->info->width,h->info->height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int rowskip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t)-h->info->width;
    tuple *row = pnm_allocpamrow(pam);
    int x;
    int y;
    if(!strcmp(pam->tuple_type,PAM_PBM_TUPLETYPE)|
       !strcmp(pam->tuple_type,PAM_PGM_TUPLETYPE))
    {
        for(y=0;y<h->info->height;++y){
            pnm_readpamrow(pam,row);
            for(x=0;x<h->info->width;++x){
                int val = 255*row[x][0]/pam->maxval;
                *dst++ = (val<<16)|
                         (val<< 8)|
                         (val<< 0)|
                         (255<<24);
            }
            dst += rowskip;
        }
    }else if(!strcmp(pam->tuple_type,PAM_PPM_TUPLETYPE)){
        for(y=0;y<h->info->height;++y){
            pnm_readpamrow(pam,row);
            for(x=0;x<h->info->width;++x){
                *dst++ = ((255*row[x][0]/pam->maxval)<<16)|
                         ((255*row[x][1]/pam->maxval)<< 8)|
                         ((255*row[x][2]/pam->maxval)<< 0)|
                         (255<<24);
            }
            dst += rowskip;
        }
    }else if(!strcmp(pam->tuple_type,PAM_PBM_ALPHA_TUPLETYPE)|
             !strcmp(pam->tuple_type,PAM_PGM_ALPHA_TUPLETYPE))
    {
        int alphamax = pam->maxval*pam->maxval;
        for(y=0;y<h->info->height;++y){
            pnm_readpamrow(pam,row);
            for(x=0;x<h->info->width;++x){
                int alpha = row[x][3];
                int val = 255*alpha*row[x][0]/alphamax;
                *dst++ = (val<<16)|
                         (val<< 8)|
                         (val<< 0)|
                         ((255*alpha/pam->maxval)<<24);
            }
            dst += rowskip;
        }
    }else if(!strcmp(pam->tuple_type,PAM_PPM_ALPHA_TUPLETYPE)){
        int alphamax = pam->maxval*pam->maxval;
        for(y=0;y<h->info->height;++y){
            pnm_readpamrow(pam,row);
            for(x=0;x<h->info->width;++x){
                int alpha = row[x][3];
                *dst++ = ((255*alpha*row[x][0]/alphamax)<<16)|
                         ((255*alpha*row[x][1]/alphamax)<< 8)|
                         ((255*alpha*row[x][2]/alphamax)<< 0)|
                         ((255*alpha/pam->maxval)<<24);
            }
            dst += rowskip;
        }
    }else{
        cairo_surface_destroy(surface);
        surface = NULL;
        h->info->error = "Unknown pixel type";
    }
    pnm_freepamrow(row);
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_netpbm_create_from_stream(abydos_plugin_handle_t *h,FILE *f)
{
    struct pam pam;
    pnm_readpaminit(f, &pam, PAM_STRUCT_SIZE(tuple_type));
    h->info->width = pam.width;
    h->info->height = pam.height;
    h->surface = _surface_from_pam(h,&pam);
    return h->surface ? 0 : -1;
}


static int
_netpbm_create_from_data(abydos_plugin_handle_t *h, const char *bytes, size_t len)
{
    FILE *f;
    int res;
    f = fmemopen((void*)bytes,len,"r");
    if(!f)
        return -1;
    res = _netpbm_create_from_stream(h,f);
    fclose(f);
    return res;
}


static int
_netpbm_create_from_file(abydos_plugin_handle_t *h, const char *filename)
{
    FILE *f;
    int res;
    f = fopen(filename,"r");
    if(!f)
        return -1;
    res = _netpbm_create_from_stream(h,f);
    fclose(f);
    return res;
}


static cairo_surface_t *
_netpbm_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *layer_visibility,int variant, double time, int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0(){
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _netpbm_get_mime_types,
        .new = _netpbm_new,
        .free = _netpbm_free,
        .create_from_data = _netpbm_create_from_data,
        .create_from_file = _netpbm_create_from_file,
        .get_image_surface = _netpbm_get_image_surface
    };
    pm_init(PACKAGE,0);
    return &p;
}
