/*
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This file is public domain.
 */
#ifndef abydos_pixels_h
#define abydos_pixels_h

#include <cairo.h>
#include <endian.h>
#include <stdint.h>

#define ABYDOS_PIXEL_FORMAT(r,g,b,a) ((r)<<24|(g)<<16|(b)<<8|(a))
#define ABYDOS_PIXEL_PREMULTIPLIED (1<<0)
#define ABYDOS_PIXEL_FLIPPED       (1<<1)

#if __BYTE_ORDER == __BIG_ENDIAN
# define ABYDOS_PIXEL_FORMAT_CAIRO_ARGB32 ABYDOS_PIXEL_FORMAT_b8g8r8a8
# define ABYDOS_PIXEL_FORMAT_CAIRO_RGB24  ABYDOS_PIXEL_FORMAT_b8g8r8
#else
# define ABYDOS_PIXEL_FORMAT_CAIRO_ARGB32 ABYDOS_PIXEL_FORMAT_a8r8g8b8
# define ABYDOS_PIXEL_FORMAT_CAIRO_RGB24  ABYDOS_PIXEL_FORMAT_x8r8g8b8
#endif

typedef union{
    int32_t id;
    struct{
#if __BYTE_ORDER == __BIG_ENDIAN
        int8_t r;
        int8_t g;
        int8_t b;
        int8_t a;
#else
        int8_t a;
        int8_t b;
        int8_t g;
        int8_t r;
#endif
    };
}channel_order_t;

typedef enum {
    ABYDOS_PIXEL_FORMAT_UNKNOWN  = -1,
    /* palette */
    ABYDOS_PIXEL_FORMAT_c8       = -2,
    ABYDOS_PIXEL_FORMAT_c8a      = -3,
    /* RGB */
    ABYDOS_PIXEL_FORMAT_r8g8b8 = ABYDOS_PIXEL_FORMAT(0,1,2,255),
    ABYDOS_PIXEL_FORMAT_x8r8g8b8 = ABYDOS_PIXEL_FORMAT(1,2,3,255),
    ABYDOS_PIXEL_FORMAT_r8g8b8a8 = ABYDOS_PIXEL_FORMAT(0,1,2,3),
    ABYDOS_PIXEL_FORMAT_a8r8g8b8 = ABYDOS_PIXEL_FORMAT(1,2,3,0),
    ABYDOS_PIXEL_FORMAT_r16g16b16 = ABYDOS_PIXEL_FORMAT(0,2,4,255),
    ABYDOS_PIXEL_FORMAT_r16g16b16a16 = ABYDOS_PIXEL_FORMAT(0,2,4,6),
    /* BGR */
    ABYDOS_PIXEL_FORMAT_b8g8r8   = ABYDOS_PIXEL_FORMAT(2,1,0,255),
    ABYDOS_PIXEL_FORMAT_x8b8g8r8 = ABYDOS_PIXEL_FORMAT(3,2,1,255),
    ABYDOS_PIXEL_FORMAT_b8g8r8a8 = ABYDOS_PIXEL_FORMAT(2,1,0,3),
    ABYDOS_PIXEL_FORMAT_a8b8g8r8 = ABYDOS_PIXEL_FORMAT(3,2,1,0),
    ABYDOS_PIXEL_FORMAT_b16g16r16a16 = ABYDOS_PIXEL_FORMAT(4,2,0,6),
    /* gray */
    ABYDOS_PIXEL_FORMAT_g8       = ABYDOS_PIXEL_FORMAT(0,0,0,255),
    ABYDOS_PIXEL_FORMAT_g8a8     = ABYDOS_PIXEL_FORMAT(0,0,0,1),
    ABYDOS_PIXEL_FORMAT_a8g8     = ABYDOS_PIXEL_FORMAT(1,1,1,0)
} abydos_pixel_format_t;

#define ABYDOS_PIXEL_FORMAT_g16 ABYDOS_PIXEL_FORMAT_g8
#define ABYDOS_PIXEL_FORMAT_g32 ABYDOS_PIXEL_FORMAT_g8

#define has_alpha(f) (ABYDOS_PIXEL_FORMAT_c8a==(f) || (((channel_order_t)f).a>=0))
#define is_gray(f) (((channel_order_t)f).r==((channel_order_t)f).g&&((channel_order_t)f).g==((channel_order_t)f).b)
#define is_16bit(f) (((channel_order_t)f).r>3 || ((channel_order_t)f).b>3)

cairo_surface_t *abydos_image_surface_create_from_pixels(abydos_pixel_format_t,int flags,const uint8_t *pixels,int width,int height,int stride,int channels,uint32_t *palette);

#endif /* abydos_pixels_h */
